from string import join,split
from xml.dom.minidom import Document
from elementtree.ElementTree import Element
from random import Random
from re import compile
from syslog import syslog, LOG_INFO, LOG_DEBUG
from time import time

class Data(object):
    '''Data is an implementation of the Restriction syntax using ElementTree for native python datatypes'''
    def __init__(self,element):
        self.element = element
        # for ease of reading
        self.attr = self.element.attrib

    def AttrMatch(self, e1, e2):
        '''AttrMatch checks if all fields in a particular pair of elements match with respect to the restriction syntax'''
        fields = e2.attrib.keys()
        if e2.attrib.get('match', None) == 'extended':
            fields.remove('match')
            for field in fields:
                (op,val) = e2.attrib[field].split(':',1)
                if op[:2] == '!d': op=op[2:]
                if op == '=':
                    if val != e1.attrib.get(field,None):
                        return False
                elif op == '!=':
                    if val == e1.attrib.get(field,None):
                        return False
                elif op == '>':
                    try:
                        if int(val) < int(e1.attrib[field]):
                            continue
                    except:
                        return False
                elif op == '<':
                    try:
                        if int(val) > int(e1.attrib[field]):
                            continue
                    except:
                        return False
                elif op == 're':
                    try:
                        r = compile(val)
                        if not r.match(e1.attrib[field]):
                            return False
                    except:
                        return False
                elif op == 'range':
                    try:
                        (min,max) = val.split('-')
                        i = int(e1.attrib.get(field,None))
                        if ((i > min) and (max > i)):
                            continue
                    except:
                        pass
                    return False
                else:
                    return False
            return True
        else:
            for field in e2.attrib.keys():
                if ((e2.attrib[field] != e1.attrib.get(field,None)) and (e2.attrib[field] != '*')):
                    return False
            return True

    def Match(self,specification):
        '''Match checks if all fields and child node fields match specification provided'''
        if self.AttrMatch(self.element, specification):
            # If we make it here, then all attributes in this element have matched
            # Now to handle children
            for child in specification.getchildren():
                found = False
                # all child nodes of the spec must be matched
                for node in self.element.findall(child.tag):
                    if self.AttrMatch(node, child):
                        found = True
                if not found:
                    return False
            return True
        else:
            return False

    def FetchElement(self, element, spec):
        '''FetchElement takes an element and a spec and produces the proper subset et Element'''
        r = Element(element.tag)
        if spec.attrib.get('match') == 'extended':
            # we need to check for '!d' operator
            fields = spec.attrib.keys()
            fields.remove('match')
            for field in fields:
                if spec.attrib[field][:2] != '!d':
                    r.attrib[field] = element.attrib.get(field, "BUG")
        else:
            for field in spec.attrib.keys():
                r.attrib[field] = element.attrib.get(field, "BUG")
        # Match against children and call FetchElement recursively
        for child in spec.getchildren():
            for node in self.element.findall(child.tag):
                if self.AttrMatch(node, child):
                    c = self.FetchElement(node, child)
                    r.append(c)
        return r

    def Fetch(self,spec=None):
        '''Fetch takes a matching spec, and produces the subset element containing only data specified'''
        if spec:
            return self.FetchElement(self.element, spec)
        return self.element

class ID(object):
    '''Autoincrementing id generator'''
    def __init__(self):
        self.id = 0

    def get(self):
        self.id += 1
        return self.id 

class RandomID(object):
    '''Somewhat randomly selected unique ID pool'''
    def __init__(self):
        self.used = []
        self.rand = Random(int(time()))

    def get(self):
        id = str(self.rand.randrange(0,2147483639)) + str(self.rand.randrange(0,2147483639))
        while id in self.used:
            id = str(self.rand.randrange(0,2147483639)) + str(self.rand.randrange(0,2147483639))
        self.used.append(id)
        return id

class DataSet(object):
    '''This class provides a list store for Data objects, complete with functions that do most of the
    restriction syntax work
    rname is the container name used in responses
    ename is the element name used for specifications
    obj is the constructor to call on new instances for the datastore
    idalloc is a unique id allocator; must be paired with self.obj function signature
    log is if add/del operations should be logged.'''
    def __init__(self,rname,ename,obj,idalloc,log):
        self.data = []
        self.rname = rname
        self.ename = ename
        self.obj = obj
        self.idalloc = idalloc
        self.log = log

    def __iter__(self):
        return iter(self.data)

    def remove(self,x):
        return self.data.remove(x)

    def append(self,x):
        return self.data.append(x)
    
    def Add(self,spec, (peer,port), callback=None):
        r = Element(self.rname)
        for s in spec.findall(self.ename):
            if self.idalloc:
                d = self.obj(s,self.idalloc.get())
            else:
                d = self.obj(s)
            if self.log:
                syslog(LOG_INFO, "Got %s from %s"%(self.ename, peer))
            self.data.append(d)
            r.append(d.Fetch(s))
            if callback: callback(d, spec.attrib)
        return r

    def Get(self,spec,(peer,port), callback=None):
        r = Element(self.rname)
        for s in spec.findall(self.ename):
            for item in [x for x in self.data if x.Match(s)]:
                r.append(item.Fetch(s))
                if callback: callback(item, spec.attrib)
        return r

    def Del(self,spec,(peer,port), callback=None):
        r = Element(self.rname)
        for s in spec.findall(self.ename):
            for item in [x for x in self.data if x.Match(s)]:
                r.append(item.Fetch(s))
                self.data.remove(item)
                if self.log:
                    syslog(LOG_INFO, "Got %s from %s"%(spec.tag,peer))
                if callback: callback(item, spec.attrib)
        return r

class SizedSet(DataSet):
    '''SizedSet is a DataSet instance that rotates through data, based on min and max'''
    def __init__(self, rname, ename, obj, idalloc, log, min, max):
        DataSet.__init__(self, rname, ename, obj, idalloc, log)
        self.min = min
        self.max = max
        
    def Add(self, spec, (peer,port), callback=None):
        r = DataSet.Add(self, spec, (peer,port), callback)
        if len(self.data) > self.max:
            self.data = self.data[-self.min:]
        return r

class SQL:
    '''sss.restriction.SQL is for classes backed by SQL tables'''
    def __init__(self,element):
        self.element = element
        # self.query%(op, fields, tables, condition)
        self.query = "%s %s from %s where %s;"
        if self.element.attrib.get("match") != 'extended':
            self.fields = self.element.attrib.keys()
            self.conditions = map(lambda x: "%s=%s"%x, self.element.attrib.iteritems())
        else:
            self.fields = [x[0] for x in self.element.attrib.iteritems() if x[1][:2] != '!d']
            self.conditions = []
            for (k,v) in self.element.attrib.iteritems():
                (op,val) = v.split(":",1)
                if op[:2] == '!d': op=op[2:]
                if op == '=':
                    self.conditions.append("%s=%s"%(k,val))
                elif op == '!=':
                    self.conditions.append("%s!=%s"%(k,val))
                elif op == '>':
                    self.conditions.append("%s > %s"%(k,val))
                elif op == '<':
                    self.conditions.append("%s < %s"%(k,val))
                elif op == 're':
                    self.conditions.append("%s REGEXP %s"%(k,val))
                elif op == 'range':
                    (min,max) = val.split('-')
                    self.conditions.append("%s > %s and %s < %s"%(min,k,k,max))
    

class restriction:
    def __init__(self,xml):
        self.XMLname=xml.tagName
        self.translations={'node-software':{'tag':"nodesoftware.node=tag.node"}}
        self.fields=[]
        self.visible=[]
        self.restrictions=[]
        self.rows=[] # These will correspond to fields
        if xml.childNodes:
            self.table=join([xml.tagName.replace('-','')]+map(lambda x:x.tagName,filter(lambda x:x.nodeType == 1,xml.childNodes)),',')
            selectionlist=[]
            for node in [xml]+xml.childNodes:
                 complexity=0
                 jcrit=[]
                 self.restrictions=[]
                 if node.hasAttribute('match'):#complexity checking
                     if (node.getAttribute('match')=='complex'):
                         complexity=1
                 for (attr,value) in node._get_attributes().items():
                     if attr!='match':
                         fullattr="%s.%s"%(node.tagName.replace('-',''),attr)
                         if value != '*':
                             self.restrictions.append((fullattr,value,complexity))
                         self.fields.append(fullattr)
                         self.visible.append(fullattr)
                 if self.translations.has_key(xml.tagName):
                     for node in xml.childNodes:
                         if self.translations[xml.tagName].has_key(node.tagName):
                             jcrit.append(self.translations[xml.tagName][node.tagName])
                 selectionlist.append(join(map(self.MakeSQL,self.restrictions)+jcrit,' and '))
                        
            self.restsql=join(selectionlist[1:len(selectionlist)],' or ')#note AND gets precedence over OR in the DB.
        else:
            self.table=xml.tagName.replace('-','')
            complexity=0
            if xml.hasAttribute('match'):#complexity checking
                if (xml.getAttribute('match')=='complex'):
                    complexity=1
            for (attr,value) in xml._get_attributes().items():
                if attr!='match':
                # Then this is a spec
		    if value != "*":
                        self.restrictions.append((attr,value,complexity))
                    self.fields.append(attr)
                    self.visible.append(attr)
            self.restsql=join(map(self.MakeSQL, self.restrictions),' and ')

    def SetRows(self,rows):
        self.rows=rows

    def GetSQL(self):
        fields=join(self.fields,',')
        if self.restsql:
            query="SELECT %s from %s where %s;"
            return query%(fields,self.table,self.restsql)
        else:
            return "SELECT %s from %s;"%(fields,self.table)

    def SetSQL(self):
        if self.restsql:
            return "UPDATE %s set %%s where %s;"%(self.table,self.restsql)
        else:
            return "UPDATE %s set %%s;"%(self.table)

    def AddSQL(self):
        query="INSERT into %s (%s) VALUES (%s);"
        fields=map(lambda x:x[0],self.restrictions)
        values=map(lambda y:"'%s'"%y,map(lambda x:x[1],self.restrictions))
        return query%(self.table,join(fields,','),join(values,','))

    def DelSQL(self):
        if self.restsql:
            query="DELETE from %s where %s;"
            return query%(self.table,self.restsql)
        else:
            return "DELETE from %s;"%(self.table)
    
    def GetXML(self):
        creator=Document()
        ret=[]
        data=self.GetFields(self.visible)
        for row in data:
            xe=creator.createElement(self.XMLname)
            for idx in xrange(len(self.visible)):
                xe.setAttribute(self.visible[idx],str(row[idx]))
                ret.append(xe)
        return ret

    def GetField(self,field):
        fidx=self.fields.index(split(field,'.')[0])
        return map(lambda x:x[fidx], self.rows)

    def GetFields(self,flist):
        idxlist=map(lambda x:self.fields.index(x),map(lambda x:split(x,'.')[0],flist))
        return map(lambda y:tuple(map(lambda x:y[x],idxlist)), self.rows)

    def MakeSQL(self,restr):
        SQL=""
        (key,value,complex) = restr
        if not complex:
            SQL = "%s = '%s'"%(key,value)
        else:
            (op,val) = value.split(':',1)
            if op == '!':
                SQL = "%s != '%s'"%(key,val)
            elif op == '>':
                SQL = "%s > '%s'"%(key,val)
            elif op == '<':
                SQL = "%s < '%s'"%(key,val)
            elif op == 'eq':
                SQL = "%s = '%s'"%(key,val)
            elif op == 're':
                SQL = "%s REGEXP '%s'"%(key,val)
        return SQL
